@@ -26,13 +26,27 @@
 #endif
 
 /* PKCS#11 Definitions for the local platform */
-#define CK_PTR *
-#define CK_DECLARE_FUNCTION(rv, func) rv func
-#define CK_DECLARE_FUNCTION_POINTER(rv, func) rv (CK_PTR func)
-#define CK_CALLBACK_FUNCTION(rv, func) CK_DECLARE_FUNCTION_POINTER(rv, func)
-#define CK_NULL_PTR ((void *) 0)
+#ifndef _WIN32
+#  define CK_PTR *
+#  define CK_DECLARE_FUNCTION(rv, func) rv func
+#  define CK_DECLARE_FUNCTION_POINTER(rv, func) rv (CK_PTR func)
+#  define CK_CALLBACK_FUNCTION(rv, func) rv (CK_PTR func)
+#  define CK_NULL_PTR ((void *) 0)
+#else
+#  define CK_PTR *
+#  define CK_DECLARE_FUNCTION(rv, func) rv __declspec(dllimport) func
+#  define CK_DECLARE_FUNCTION_POINTER(rv, func) rv __declspec(dllimport) (CK_PTR func)
+#  define CK_CALLBACK_FUNCTION(rv, func) rv (CK_PTR func)
+#  define CK_NULL_PTR ((void *) 0)
+#  pragma pack(push, cryptoki, 1)
+#endif
+
 #include "pkcs11.h"
 
+#ifdef _WIN32
+#  pragma pack(pop, cryptoki)
+#endif
+
 struct tclpkcs11_interpdata {
 	/* Handle Hash Table */
 	Tcl_HashTable handles;
@@ -45,6 +59,7 @@ struct tclpkcs11_handle {
 	CK_FUNCTION_LIST_PTR pkcs11;
 
 	/* Session Management */
+	int session_active;
 	CK_SLOT_ID session_slot;
 	CK_SESSION_HANDLE session;
 };
@@ -242,7 +257,7 @@ MODULE_SCOPE Tcl_Obj *tclpkcs11_bytearray_to_string(const unsigned char *data, u
 	char buf[1024];
 
 	if (data == NULL) {
-		return(retval);
+		return(Tcl_NewObj());
 	}
 
 	for (bufidx = idx = 0; (idx < datalen) && (bufidx < sizeof(buf)); idx++) {
@@ -250,7 +265,7 @@ MODULE_SCOPE Tcl_Obj *tclpkcs11_bytearray_to_string(const unsigned char *data, u
 		buf[bufidx++] = alphabet[data[idx] & 0xf];
 	}
 
-	retval = Tcl_NewByteArrayObj(buf, bufidx);
+	retval = Tcl_NewByteArrayObj((unsigned char *) buf, bufidx);
 
 	return(retval);
 }
@@ -353,15 +368,14 @@ MODULE_SCOPE int tclpkcs11_start_session(struct tclpkcs11_handle *handle, CK_SLO
 	CK_SESSION_HANDLE tmp_session;
 	CK_RV chk_rv;
 
-	if (handle->session != -1) {
+	if (handle->session_active) {
 		if (handle->session_slot == slot) {
 			return(CKR_OK);
 		}
 
 		/* Close the existing session and create a new one */
+		handle->session_active = 0;
 		chk_rv = handle->pkcs11->C_CloseSession(handle->session);
-		handle->session = -1;
-		handle->session_slot = -1;
 		if (chk_rv != CKR_OK) {
 			return(chk_rv);
 		}
@@ -370,14 +384,13 @@ MODULE_SCOPE int tclpkcs11_start_session(struct tclpkcs11_handle *handle, CK_SLO
 	chk_rv = handle->pkcs11->C_OpenSession(slot, CKF_SERIAL_SESSION, NULL, NULL, &tmp_session);
 	if (chk_rv != CKR_OK) {
 		handle->pkcs11->C_CloseSession(handle->session);
-		handle->session = -1;
-		handle->session_slot = -1;
 
 		return(chk_rv);
 	}
 
 	handle->session = tmp_session;
 	handle->session_slot = slot;
+	handle->session_active = 1;
 
 	return(CKR_OK);
 }
@@ -385,10 +398,9 @@ MODULE_SCOPE int tclpkcs11_start_session(struct tclpkcs11_handle *handle, CK_SLO
 MODULE_SCOPE int tclpkcs11_close_session(struct tclpkcs11_handle *handle) {
 	CK_RV chk_rv;
 
-	if (handle->session != -1) {
+	if (handle->session_active) {
+		handle->session_active = 0;
 		chk_rv = handle->pkcs11->C_CloseSession(handle->session);
-		handle->session = -1;
-		handle->session_slot = -1;
 
 		if (chk_rv != CKR_OK) {
 			return(chk_rv);
@@ -482,8 +494,8 @@ MODULE_SCOPE int tclpkcs11_load_module(ClientData cd, Tcl_Interp *interp, int ob
 	int is_new_entry;
 
 	CK_C_INITIALIZE_ARGS initargs;
-	CK_RV (*getFuncList)(CK_FUNCTION_LIST_PTR_PTR ppFunctionList);
-	CK_FUNCTION_LIST_PTR pkcs11_function_list;
+	CK_C_GetFunctionList getFuncList;
+	CK_FUNCTION_LIST_PTR pkcs11_function_list = NULL;
 	CK_RV chk_rv;
 
 	if (!cd) {
@@ -572,8 +584,7 @@ MODULE_SCOPE int tclpkcs11_load_module(ClientData cd, Tcl_Interp *interp, int ob
 	/* Initialize the per-handle structure */
 	new_handle->base = handle;
 	new_handle->pkcs11 = pkcs11_function_list;
-	new_handle->session = -1;
-	new_handle->session_slot = -1;
+	new_handle->session_active = 0;
 
 	Tcl_SetHashValue(tcl_handle_entry, (ClientData) new_handle);
 
@@ -932,6 +943,7 @@ MODULE_SCOPE int tclpkcs11_list_certs(ClientData cd, Tcl_Interp *interp, int obj
 			}
 
 			curr_attr->pValue = NULL;
+			curr_attr->ulValueLen = 0;
 		}
 
 		/* Determine size of values to allocate */
@@ -1207,9 +1219,15 @@ MODULE_SCOPE int tclpkcs11_logout(ClientData cd, Tcl_Interp *interp, int objc, T
 
 	chk_rv = handle->pkcs11->C_Logout(handle->session);
 	if (chk_rv != CKR_OK) {
-		Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
+		if (chk_rv == CKR_DEVICE_REMOVED) {
+			handle->session_active = 0;
 
-		return(TCL_ERROR);
+			handle->pkcs11->C_CloseSession(handle->session);
+		} else {
+			Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
+
+			return(TCL_ERROR);
+		}
 	}
 
 	Tcl_SetObjResult(interp, Tcl_NewBooleanObj(1));
@@ -1232,6 +1250,7 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 	int tcl_keylist_llength, idx;
 	int input_len;
 	CK_ULONG resultbuf_len;
+	int sign;
 	int tcl_rv;
 
 	CK_SLOT_ID slotid;
@@ -1419,7 +1438,7 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 	handle->pkcs11->C_FindObjectsFinal(handle->session);
 
 	if (foundObjs < 1) {
-		Tcl_SetObjResult(interp, Tcl_NewStringObj("unable to find private key that cooresponds to this certificate", -1));
+		Tcl_SetObjResult(interp, Tcl_NewStringObj("PKCS11_ERROR MAYBE_LOGIN", -1));
 
 		return(TCL_ERROR);
 	}
@@ -1427,15 +1446,45 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 	/* Perform the PKI operation (encrypt/decrypt) */
 	input = Tcl_GetByteArrayFromObj(tcl_input, &input_len);
 	if (encrypt) {
-		chk_rv = handle->pkcs11->C_SignInit(handle->session, &mechanism, hObject);
+		sign = 0;
+		chk_rv = handle->pkcs11->C_EncryptInit(handle->session, &mechanism, hObject);
 		if (chk_rv != CKR_OK) {
-			Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
+			if (chk_rv == CKR_FUNCTION_NOT_SUPPORTED) {
+				sign = 1;
+				chk_rv = handle->pkcs11->C_SignInit(handle->session, &mechanism, hObject);
+				if (chk_rv != CKR_OK) {
+					Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
 
-			return(TCL_ERROR);
+					return(TCL_ERROR);
+				}
+			}
 		}
 
 		resultbuf_len = sizeof(resultbuf);
-		chk_rv = handle->pkcs11->C_Sign(handle->session, input, input_len, resultbuf, &resultbuf_len);
+		if (!sign) {
+			chk_rv = handle->pkcs11->C_Encrypt(handle->session, input, input_len, resultbuf, &resultbuf_len);
+		} else {
+			/* Some PKCS#11 drivers will not accept pre-padded input, so we must unpad it here */
+			if (input_len > 3) {
+				if (input[0] == 0x00 && input[1] == 0x01) {
+					input = input + 2;
+					input_len -= 2;
+					while (*input == 0xff && input_len > 0) {
+						input++;
+						input_len--;
+					}
+
+					if (input_len) {
+						if (input[0] == 0x00) {
+							input++;
+							input_len--;
+						}
+					}
+				}
+			}
+
+			chk_rv = handle->pkcs11->C_Sign(handle->session, input, input_len, resultbuf, &resultbuf_len);
+		}
 		if (chk_rv != CKR_OK) {
 			if (chk_rv == CKR_BUFFER_TOO_SMALL) {
 				/* Terminate decryption operation */
@@ -1483,6 +1532,40 @@ MODULE_SCOPE int tclpkcs11_decrypt(ClientData cd, Tcl_Interp *interp, int objc,
 	return(tclpkcs11_perform_pki(0, cd, interp, objc, objv));
 }
 
+MODULE_SCOPE void tclpkcs11_unloadall(ClientData cd) {
+	struct tclpkcs11_interpdata *interpdata;
+	struct tclpkcs11_handle *handle;
+	Tcl_HashEntry *tcl_handle_entry;
+	Tcl_HashSearch search;
+
+	if (!cd) {
+		return;
+	}
+
+	interpdata = (struct tclpkcs11_interpdata *) cd;
+
+	for (
+		tcl_handle_entry = Tcl_FirstHashEntry(&interpdata->handles, &search);
+		tcl_handle_entry;
+		tcl_handle_entry = Tcl_NextHashEntry(&search)
+	) {
+
+		handle = (struct tclpkcs11_handle *) Tcl_GetHashValue(tcl_handle_entry);
+
+		if (handle->pkcs11 && handle->pkcs11->C_Finalize) {
+			handle->pkcs11->C_Finalize(NULL);
+		}
+
+		tclpkcs11_int_unload_module(handle->base);
+
+		ckfree((char *) handle);
+
+	}
+
+	return;
+}
+
+
 /*
  * Tcl Loadable Module Initialization
  */
@@ -1555,6 +1638,9 @@ int Tclpkcs11_Init(Tcl_Interp *interp) {
 		return(TCL_ERROR);
 	}
 
+	/* Create an exit handler to unload and close all PKCS#11 modules */
+	Tcl_CreateExitHandler(tclpkcs11_unloadall, interpdata);
+
 	/* Register PKI handlers */
 	Tcl_ObjSetVar2(interp,
 	               Tcl_NewStringObj("pki::handlers", -1),
